//! This struct defines presets.

use crate::{
	iotro::{Language, BRITISH, GERMAN},
	project::Resolution
};
use anyhow::bail;
use serde::{Deserialize, Serialize};
use serde_with::{serde_as, DisplayFromStr};
use std::{fs, io};

#[serde_as]
#[derive(Deserialize, Serialize)]
pub struct Preset {
	// options for the intro slide
	pub course: String,
	pub label: String,
	pub docent: String,

	/// Course language
	#[serde(default)]
	#[serde_as(as = "DisplayFromStr")]
	pub lang: Language<'static>,

	// coding options
	#[serde_as(as = "DisplayFromStr")]
	pub transcode_start: Resolution,
	#[serde_as(as = "Option<DisplayFromStr>")]
	pub transcode: Option<Resolution>
}

pub fn preset_23ws_malo2() -> Preset {
	Preset {
		course: "23ws-malo2".into(),
		label: "Mathematische Logik II".into(),
		docent: "Prof. E. Grädel".into(),
		lang: GERMAN,
		transcode_start: "1440p".parse().unwrap(),
		transcode: Some("360p".parse().unwrap())
	}
}

pub fn preset_24ss_algomod() -> Preset {
	Preset {
		course: "24ss-algomod".into(),
		label: "Algorithmische Modelltheorie".into(),
		docent: "Prof. E. Grädel".into(),
		lang: GERMAN,
		transcode_start: "1440p".parse().unwrap(),
		transcode: Some("720p".parse().unwrap())
	}
}

pub fn preset_24ss_qc() -> Preset {
	Preset {
		course: "24ss-qc".into(),
		label: "Introduction to Quantum Computing".into(),
		docent: "Prof. D. Unruh".into(),
		lang: BRITISH,
		transcode_start: "1440p".parse().unwrap(),
		transcode: Some("720p".parse().unwrap())
	}
}

impl Preset {
	pub fn find(name: &str) -> anyhow::Result<Self> {
		match fs::read(name) {
			Ok(buf) => return Ok(toml::from_slice(&buf)?),
			Err(err) if err.kind() == io::ErrorKind::NotFound => {},
			Err(err) => return Err(err.into())
		}
		Ok(match name {
			"23ws-malo2" => preset_23ws_malo2(),
			"24ss-algomod" => preset_24ss_algomod(),
			"24ss-qc" => preset_24ss_qc(),
			_ => bail!("Unknown preset {name:?}")
		})
	}
}
